/*  Canon Electronics Scanner Driver for Linux Version 1.00
 *  Copyright CANON ELECTRONICS INC. 2019
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include "sane/sane.h"
#include "sane/sanei_usb.h"
#include "sane/sanei_config.h"

//#define CU_DEBUG
#ifdef CU_DEBUG
#define CU_LOG_FILE     "/tmp/Canon_DR_DRIVERCONTROL.txt"
#endif

#define NEW_PROTOCOL_VERSION
#define USE_USB_CLEAR_HALT

/*------------------------------------------------------------------------------*/

#ifdef CU_DEBUG
FILE* CU_LOG_FP ;
#endif
#ifdef CU_DEBUG
#define CU_PRINT(...) \
{\
           if(!CU_LOG_FP)\
                        CU_LOG_FP = fopen( CU_LOG_FILE, "a" );\
           fprintf(CU_LOG_FP,"[%s,%d,%s]",__FILE__,__LINE__,__FUNCTION__);\
           fprintf(CU_LOG_FP,__VA_ARGS__);\
           fprintf(CU_LOG_FP,"\n");\
           fflush(CU_LOG_FP);\
}
#else
#define  CU_PRINT(fmt,...) \
{}
#endif/* end of #ifdef CU_DEBUG */

#ifdef CU_DEBUG
#define CU_PRINT_BYTES(buf, cnt)\
{\
        if(!CU_LOG_FP)\
                CU_LOG_FP = fopen( CU_LOG_FILE, "w" );\
        int i = 0,tmp = 0;\
        fprintf(CU_LOG_FP,"-----%s-----%d-----%s-----\n",__FILE__,__LINE__,__FUNCTION__);\
        fprintf(CU_LOG_FP,"The bytes length is %ld \n", (int)cnt);\
        if((int)cnt>100){\
                fprintf(CU_LOG_FP,"The bytes are too long, only print the first 100 bytes");\
        }\
        for(i = 0;i<((int)cnt<100? (int)cnt:100);i++){\
                if(i%16 == 0)\
                        fprintf(CU_LOG_FP,"\n");\
                tmp = buf[i];\
                tmp = tmp & 0xFF;\
                fprintf(CU_LOG_FP," %.2x ",tmp);\
        }\
        fprintf(CU_LOG_FP,"\n");\
        fflush(CU_LOG_FP);\
}
#else
#define CU_PRINT_BYTES(buf,cnt) \
{}
#endif/* end of #ifdef CU_DEBUG */

/*------------------------------------------------------------------------------*/


struct DeviceID
{
	SANE_Int g_nVendor;
	SANE_Int g_nProduct;
};
static struct DeviceID g_DeviceID[] = {
	{0x1083, 0x166e},	/* G2140 */
	{0x1083, 0x166f},	/* G2110 */
	{0x1083, 0x1672},	/* G2090 */
	{0,0},
};

static int usbFd = -1;
static char *deviceName = NULL;
struct UserData
{
	char enableSeparationForceControl;
	char enableManualFeedControl;
	char enableManualFeedRollerRotatingTimeControl;
	char enableStapleDetectionControl;
	char enableLongDocumentControl;
	char enableSlowFeedAtLongDocumentControl;
	char enableTrayPositionControl;
	char enableFeedIntervalControl;
	char enableDustDetetionControl;
};
#define BITON(data, bit)	((data) |= (bit))
#define BITOFF(data, bit)	((data) &= ~(bit))

#if !defined(USE_USB_CLEAR_HALT)
static SANE_Status _usbClear(int _usbFd, char *_deviceName)
{
	sanei_usb_close(_usbFd);
	return sanei_usb_open(_deviceName, &_usbFd);
}
#define usbClear(fd, name)	_usbClear(fd)
#else
static SANE_Status _usbClear2(int _usbFd)
{
	return sanei_usb_clear_halt(_usbFd);
}
#define usbClear(fd, name)	_usbClear2(fd)
#endif

SANE_Status ExecRead(SANE_Byte *cdb, SANE_Word cdb_size, SANE_Byte *data, SANE_Word data_size)
{
        SANE_Status status = SANE_STATUS_GOOD;
        SANE_Word n=cdb_size;

        status = sanei_usb_write_bulk(usbFd, cdb, (size_t *)&n);
	CU_PRINT("sanei_usb_write_bulk()");
        if (status!=SANE_STATUS_GOOD) {
		CU_PRINT("sanei_usb_write_bulk(cdb) error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=cdb_size) {
                CU_PRINT("n!=cdb_size error n is %d, cdb_size is  %d", n, cdb_size);
                return SANE_STATUS_GOOD;
        }

#if defined(NEW_PROTOCOL_VERSION)
	{
		int bContinue = 1;
		SANE_Byte resp[8]={0};
		while (bContinue) {
			n = sizeof(resp);
        		status = sanei_usb_read_bulk(usbFd, resp, (size_t *)&n);
        		CU_PRINT("sanei_usb_read_bulk()");
        		if (status!=SANE_STATUS_GOOD) {
                		CU_PRINT("sanei_usb_read_bulk(data) error %d", status);
				status = usbClear(usbFd, deviceName);
                		return status;
        		}
        		if (n!=sizeof(resp)) {
                		CU_PRINT("n!=data_size error n is %d, data_size is  %d", n, sizeof(resp));
                		return SANE_STATUS_GOOD;
        		}

			uint32_t s = (((uint32_t)resp[0] & 0xff) << 24) | \
				(((uint32_t)resp[1] & 0xff) << 16) | \
				(((uint32_t)resp[2] & 0xff) << 8) | \
				(((uint32_t)resp[3] & 0xff) << 0);
			switch (s) {
			case 0:
				bContinue = 0;
				break;
			case 1:
				break;
			default:
				if (s & 0x80) {
					CU_PRINT("error(%d)", s & 0x7f);
					return SANE_STATUS_INVAL;
				}
			}
		}
	}
#endif	//defined(NEW_PROTOCOL_VERSION)

        n = data_size;
        status = sanei_usb_read_bulk(usbFd, data, (size_t *)&n);
        CU_PRINT("sanei_usb_read_bulk()");
        if (status!=SANE_STATUS_GOOD) {
                CU_PRINT("sanei_usb_read_bulk(data) error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=data_size) {
                CU_PRINT("n!=data_size error n is %d, data_size is  %d", n, data_size);
                return SANE_STATUS_GOOD;
        }

#if defined(NEW_PROTOCOL_VERSION)
	{
		int bContinue = 1;
		SANE_Byte resp[8]={0};
		while (bContinue) {
			n = sizeof(resp);
        		status = sanei_usb_read_bulk(usbFd, resp, (size_t *)&n);
        		CU_PRINT("sanei_usb_read_bulk()");
        		if (status!=SANE_STATUS_GOOD) {
                		CU_PRINT("sanei_usb_read_bulk(data) error %d", status);
				status = usbClear(usbFd, deviceName);
                		return status;
        		}
        		if (n!=sizeof(resp)) {
                		CU_PRINT("n!=data_size error n is %d, data_size is  %d", n, sizeof(resp));
                		return SANE_STATUS_GOOD;
        		}

			uint32_t s = (((uint32_t)resp[0] & 0xff) << 24) | \
				(((uint32_t)resp[1] & 0xff) << 16) | \
				(((uint32_t)resp[2] & 0xff) << 8) | \
				(((uint32_t)resp[3] & 0xff) << 0);
			switch (s) {
			case 0:
				bContinue = 0;
				break;
			case 1:
				break;
			default:
				if (s & 0x80) {
					CU_PRINT("error(%d)", s & 0x7f);
					return SANE_STATUS_INVAL;
				}
			}
		}
	}
#else	//!defined(NEW_PROTOCOL_VERSION)
        SANE_Byte scanner_status[4]={0};
        n=sizeof(scanner_status);
        status = sanei_usb_read_bulk(usbFd, scanner_status, (size_t *)&n);
        CU_PRINT("sanei_usb_read_bulk()");
        if (status!=SANE_STATUS_GOOD) {
                CU_PRINT("sanei_usb_read_bulk(scanner_status) error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=sizeof(scanner_status)) {
                CU_PRINT("n!=1 error n is %d", n);
                return SANE_STATUS_GOOD;
        }
#endif	//defined(NEW_PROTOCOL_VERSION)
        return SANE_STATUS_GOOD;
}

SANE_Status ExecWrite(SANE_Byte *cdb, long cdb_size, SANE_Byte *data, long data_size)
{
        SANE_Status status = SANE_STATUS_GOOD;
        SANE_Word n=cdb_size;
        status = sanei_usb_write_bulk(usbFd, cdb, (size_t *)&n);
        CU_PRINT("sanei_usb_write_bulk()");
        if (status!=SANE_STATUS_GOOD) {
                CU_PRINT("sanei_usb_write_bulk(cdb) error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=cdb_size) {
                CU_PRINT("n!=cdb_size error n is %d, cdb_size is %ld", n, cdb_size);
                return SANE_STATUS_GOOD;
        }

#if defined(NEW_PROTOCOL_VERSION)
	{
		int bContinue = 1;
		SANE_Byte resp[8]={0};
		while (bContinue) {
			n = sizeof(resp);
        		status = sanei_usb_read_bulk(usbFd, resp, (size_t *)&n);
        		CU_PRINT("sanei_usb_read_bulk()");
        		if (status!=SANE_STATUS_GOOD) {
                		CU_PRINT("sanei_usb_read_bulk(data) error %d", status);
				status = usbClear(usbFd, deviceName);
                		return status;
        		}
        		if (n!=sizeof(resp)) {
                		CU_PRINT("n!=data_size error n is %d, data_size is  %d", n, sizeof(resp));
                		return SANE_STATUS_GOOD;
        		}

			uint32_t s = (((uint32_t)resp[0] & 0xff) << 24) | \
				(((uint32_t)resp[1] & 0xff) << 16) | \
				(((uint32_t)resp[2] & 0xff) << 8) | \
				(((uint32_t)resp[3] & 0xff) << 0);
			switch (s) {
			case 0:
				bContinue = 0;
				break;
			case 1:
				break;
			default:
				if (s & 0x80) {
					CU_PRINT("error(%d)", s & 0x7f);
					return SANE_STATUS_INVAL;
				}
			}
		}
	}
#endif	//defined(NEW_PROTOCOL_VERSION)

        n = data_size;
        status = sanei_usb_write_bulk(usbFd, data, (size_t *)&n);
        CU_PRINT("sanei_usb_write_bulk()");
        if (status!=SANE_STATUS_GOOD) {
                CU_PRINT("sanei_usb_write_bulk(data) error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=data_size) {
                CU_PRINT("n!=data_size error n is %d, data_size is %ld", n, data_size);
                return SANE_STATUS_GOOD;
        }

#if defined(NEW_PROTOCOL_VERSION)
	{
		int bContinue = 1;
		SANE_Byte resp[8]={0};
		while (bContinue) {
			n = sizeof(resp);
        		status = sanei_usb_read_bulk(usbFd, resp, (size_t *)&n);
        		CU_PRINT("sanei_usb_read_bulk()");
        		if (status!=SANE_STATUS_GOOD) {
                		CU_PRINT("sanei_usb_read_bulk(data) error %d", status);
				status = usbClear(usbFd, deviceName);
                		return status;
        		}
        		if (n!=sizeof(resp)) {
                		CU_PRINT("n!=data_size error n is %d, data_size is  %d", n, sizeof(resp));
                		return SANE_STATUS_GOOD;
        		}

			uint32_t s = (((uint32_t)resp[0] & 0xff) << 24) | \
				(((uint32_t)resp[1] & 0xff) << 16) | \
				(((uint32_t)resp[2] & 0xff) << 8) | \
				(((uint32_t)resp[3] & 0xff) << 0);
CU_PRINT("s = %x", s);
			switch (s) {
			case 0:
				bContinue = 0;
				break;
			case 1:
				break;
			default:
				if (s & 0x80) {
					CU_PRINT("error(%d)", s & 0x7f);
					return SANE_STATUS_INVAL;
				}
			}
		}
	}
#else	//!defined(NEW_PROTOCOL_VERSION)
        SANE_Byte scanner_status[4]={0};
        n=sizeof(scanner_status);
        status = sanei_usb_read_bulk(usbFd, scanner_status, (size_t *)&n);
        CU_PRINT("sanei_usb_read_bulk()");
        if (status!=SANE_STATUS_GOOD) {
                CU_PRINT("sanei_usb_read_bulk(scanner_status) error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=sizeof(scanner_status)) {
                CU_PRINT("n!=1 error n is %d", n);
                return SANE_STATUS_GOOD;
        }
#endif	//defined(NEW_PROTOCOL_VERSION)
        return SANE_STATUS_GOOD;
}

SANE_Status ExecNone(SANE_Byte *cdb, long cdb_size)
{
        SANE_Status status = SANE_STATUS_GOOD;
        SANE_Word n=cdb_size;
        status = sanei_usb_write_bulk(usbFd, cdb, (size_t *)&n);
        CU_PRINT("sanei_usb_write_bulk()");
        if (status!=SANE_STATUS_GOOD) {
                CU_PRINT("sanei_usb_write_bulk(cdb) error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=cdb_size) {
                CU_PRINT("n!=cdb_size error n is %d, cdb_size is %ld", n, cdb_size);
                return SANE_STATUS_GOOD;
        }

#if defined(NEW_PROTOCOL_VERSION)
	{
		int bContinue = 1;
		SANE_Byte resp[8]={0};
		while (bContinue) {
			n = sizeof(resp);
        		status = sanei_usb_read_bulk(usbFd, resp, (size_t *)&n);
        		CU_PRINT("sanei_usb_read_bulk()");
        		if (status!=SANE_STATUS_GOOD) {
                		CU_PRINT("sanei_usb_read_bulk(data) error %d", status);
				status = usbClear(usbFd, deviceName);
                		return status;
        		}
        		if (n!=sizeof(resp)) {
                		CU_PRINT("n!=data_size error n is %d, data_size is  %d", n, sizeof(resp));
                		return SANE_STATUS_GOOD;
        		}

			uint32_t s = (((uint32_t)resp[0] & 0xff) << 24) | \
				(((uint32_t)resp[1] & 0xff) << 16) | \
				(((uint32_t)resp[2] & 0xff) << 8) | \
				(((uint32_t)resp[3] & 0xff) << 0);
			switch (s) {
			case 0:
				bContinue = 0;
				break;
			case 1:
				break;
			default:
				if (s & 0x80) {
					CU_PRINT("error(%d)", s & 0x7f);
					return SANE_STATUS_INVAL;
				}
			}
		}
	}
#endif	//defined(NEW_PROTOCOL_VERSION)

#if defined(NEW_PROTOCOL_VERSION)
	{
		int bContinue = 1;
		SANE_Byte resp[8]={0};
		while (bContinue) {
			n = sizeof(resp);
        		status = sanei_usb_read_bulk(usbFd, resp, (size_t *)&n);
        		CU_PRINT("sanei_usb_read_bulk()");
        		if (status!=SANE_STATUS_GOOD) {
                		CU_PRINT("sanei_usb_read_bulk(data) error %d", status);
				status = usbClear(usbFd, deviceName);
                		return status;
        		}
        		if (n!=sizeof(resp)) {
                		CU_PRINT("n!=data_size error n is %d, data_size is  %d", n, sizeof(resp));
                		return SANE_STATUS_GOOD;
        		}

			uint32_t s = (((uint32_t)resp[0] & 0xff) << 24) | \
				(((uint32_t)resp[1] & 0xff) << 16) | \
				(((uint32_t)resp[2] & 0xff) << 8) | \
				(((uint32_t)resp[3] & 0xff) << 0);
			switch (s) {
			case 0:
				bContinue = 0;
				break;
			case 1:
				break;
			default:
				if (s & 0x80) {
					CU_PRINT("error(%d)", s & 0x7f);
					return SANE_STATUS_INVAL;
				}
			}
		}
	}
#else	//!defined(NEW_PROTOCOL_VERSION)
        SANE_Byte scanner_status[4]={0};
        n=sizeof(scanner_status);
        status = sanei_usb_read_bulk(usbFd, scanner_status, (size_t *)&n);
        CU_PRINT("sanei_usb_read_bulk()");
        if (status!=SANE_STATUS_GOOD) {
                CU_PRINT("sanei_usb_read_bulk() error %d", status);
		status = usbClear(usbFd, deviceName);
                return status;
        }
        if (n!=sizeof(scanner_status)) {
                CU_PRINT("n!=1 error n is %d", n);
                return SANE_STATUS_GOOD;
        }
#endif	//defined(NEW_PROTOCOL_VERSION)
        return SANE_STATUS_GOOD;
}

static SANE_Status check_present (SANE_String_Const devname)
{
	deviceName = (char *)devname;
	CU_PRINT("check_present: %s detected.", devname);
	return SANE_STATUS_GOOD;
}

void testunitready()
{
	SANE_Byte cdb[24] = {0x00, 0x00, 0x00, 0x14, 0x00, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
        SANE_Status status;

        CU_PRINT("test unit ready start");
        status = ExecNone(cdb, sizeof(cdb));
        if (status!=SANE_STATUS_GOOD) return;

	CU_PRINT_BYTES(cdb, sizeof(cdb));
        CU_PRINT("test unit ready end");
}

void inquiry()
{
	SANE_Byte cdb[24] = {0x00, 0x00, 0x00, 0x14, 0x00, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x12, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
        SANE_Byte data[64]={0};
        SANE_Status status;

        CU_PRINT("inquiry start");
        status = ExecRead(cdb, sizeof(cdb), data, sizeof(data));
        if (status!=SANE_STATUS_GOOD) return;

	CU_PRINT_BYTES(cdb, sizeof(cdb));
	CU_PRINT_BYTES(data, sizeof(data));
        CU_PRINT("inquiry end");
}

void requestsense(SANE_Byte *data, int length)
{
	SANE_Byte cdb[24] = {0x00, 0x00, 0x00, 0x14, 0x00, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x03, 0x00, 0x00, 0x00, 0x0e, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
        SANE_Byte _data[14]={0}, _length=14;
        SANE_Status status;

	if (!data) {
		data = _data;
		length = _length;
	}

        CU_PRINT("requestsense start");
        status = ExecRead(cdb, sizeof(cdb), data, length);
        if (status!=SANE_STATUS_GOOD) return;

	CU_PRINT_BYTES(cdb, sizeof(cdb));
	CU_PRINT_BYTES(data, length);
        CU_PRINT("requestsense end");
}

void readuserdata(SANE_Byte *data, int length)
{
	SANE_Byte readcdb[24] = {0x00, 0x00, 0x00, 0x14, 0x00, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x28, 0x00, 0x8c, 0x00, 0x00, 0x01, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00};
        SANE_Byte _data[128]={0}, _length=128;
        SANE_Status status;

	if (!data) {
		data = _data;
		length = _length;
	}

        CU_PRINT("readuserdata start");
        status = ExecRead(readcdb, sizeof(readcdb), data, length);
        if (status!=SANE_STATUS_GOOD) {
		requestsense(NULL, 0);
		return;
	}

	CU_PRINT_BYTES(readcdb, sizeof(readcdb));
	CU_PRINT_BYTES(data, length);
        CU_PRINT("readuserdata end");
}

void senduserdata(SANE_Byte *_data, int _length)
{
	SANE_Byte sendcdb[24] = {0x00, 0x00, 0x00, 0x14, 0x00, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 
		0x2a, 0x00, 0x8c, 0x00, 0x00, 0x01, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00};
        SANE_Byte data[12+128]={0x00, 0x00, 0x00, 0x88, 0x00, 0x02, 0xb0, 0x00, 0x00, 0x00, 0x00, 0x00};
	int length=12+128;
        SANE_Status status;

	if (_data) {
		memcpy(&data[12], _data, _length < 128 ? _length : 128);
	}

        CU_PRINT("senduserdata start");
        status = ExecWrite(sendcdb, sizeof(sendcdb), data, length);
        if (status!=SANE_STATUS_GOOD) {
		requestsense(NULL, 0);
		return;
	}

	CU_PRINT_BYTES(sendcdb, sizeof(sendcdb));
	CU_PRINT_BYTES(data, length);
        CU_PRINT("senduserdata end");
}

void usage(char *name)
{
			fprintf(stderr, 
"\n\
Usage: %s [REQUIRES]...\n\
Start scanner control.\n\
Set scanner settings to a scanner.\n\
\n\
Parameters are separated by blank from single character requires.\n\
-a (y|n)		Enable all below settings on scanner.\n\
-f (y|n)		Enable separation force control on scanner.\n\
-m (y|n)		Enable manual feed control on scanner.\n\
-r (y|n)		Enable manual feed roller rotating time control on scanner.\n\
-s (y|n)		Enable staple detection control on scanner.\n\
-l (y|n)		Enable long document control on scanner.\n\
-o (y|n)		Enable slow feed at long document control on scanner.\n\
-t (y|n)		Enable tray position control on scanner.\n\
-i (y|n)		Enable feed interval control on scanner.\n\
-d (y|n)		Enable dust detection control on scanner.\n\
",
			name);
}

int main(int argc, char **argv)
{
	int opt, i;
	const char *optstring = "f:m:r:s:l:o:t:i:d:a:";
	SANE_Status status = SANE_STATUS_GOOD;
	char enableAllUserDataControl;
	struct UserData userdata;

	enableAllUserDataControl = 0;
	memset(&userdata, 0, sizeof(userdata));

	while ((opt = getopt(argc, argv, optstring)) != -1) {
		switch (opt) {
		case 'a':	//all control */
			if (!strcmp(optarg, "y")) {
				enableAllUserDataControl++;
			} else if (!strcmp(optarg, "n")) {
				enableAllUserDataControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'f':	/* enable separation force control */
			if (!strcmp(optarg, "y")) {
				userdata.enableSeparationForceControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableSeparationForceControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'm':	/* enable manual feed control */
			if (!strcmp(optarg, "y")) {
				userdata.enableManualFeedControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableManualFeedControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'r':	/* enable manual feed roller rotating time control */
			if (!strcmp(optarg, "y")) {
				userdata.enableManualFeedRollerRotatingTimeControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableManualFeedRollerRotatingTimeControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 's':	/* enable staple detection control */
			if (!strcmp(optarg, "y")) {
				userdata.enableStapleDetectionControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableStapleDetectionControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'l':	/* enable long document control */
			if (!strcmp(optarg, "y")) {
				userdata.enableLongDocumentControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableLongDocumentControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'o':	/* enable slow feed at long document control */
			if (!strcmp(optarg, "y")) {
				userdata.enableSlowFeedAtLongDocumentControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableSlowFeedAtLongDocumentControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 't':	/* enable tray position control */
			if (!strcmp(optarg, "y")) {
				userdata.enableTrayPositionControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableTrayPositionControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'i':	/* enable feed interval control */
			if (!strcmp(optarg, "y")) {
				userdata.enableFeedIntervalControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableFeedIntervalControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		case 'd':	/* enable dust detection control */
			if (!strcmp(optarg, "y")) {
				userdata.enableDustDetetionControl++;
			} else if (!strcmp(optarg, "n")) {
				userdata.enableDustDetetionControl--;
			} else {
				usage(argv[0]);
				exit(EXIT_FAILURE);
			}
			break;
		default:
			usage(argv[0]);
			exit(EXIT_FAILURE);
		}
	}

	if (1 == optind) {
		usage(argv[0]);
		exit(EXIT_SUCCESS);
	}
	if (optind < argc) {
		fprintf(stderr, "%s, argument without option: %s!\r\n", argv[0], argv[argc - 1]);
		CU_PRINT("%s, argument without option: %s!", argv[0], argv[argc - 1]);
		usage(argv[0]);
		exit(EXIT_FAILURE);
	}

	if (enableAllUserDataControl) {
		memset(&userdata, 0, sizeof(userdata));
		if (0 < enableAllUserDataControl) {
			userdata.enableSeparationForceControl++;
			userdata.enableManualFeedControl++;
			userdata.enableManualFeedRollerRotatingTimeControl++;
			userdata.enableStapleDetectionControl++;
			userdata.enableLongDocumentControl++;
			userdata.enableSlowFeedAtLongDocumentControl++;
			userdata.enableTrayPositionControl++;
			userdata.enableFeedIntervalControl++;
			userdata.enableDustDetetionControl++;
		} else {	//if (0 < enableAllUserDataControl)
			userdata.enableSeparationForceControl--;
			userdata.enableManualFeedControl--;
			userdata.enableManualFeedRollerRotatingTimeControl--;
			userdata.enableStapleDetectionControl--;
			userdata.enableLongDocumentControl--;
			userdata.enableSlowFeedAtLongDocumentControl--;
			userdata.enableTrayPositionControl--;
			userdata.enableFeedIntervalControl--;
			userdata.enableDustDetetionControl--;
		}
	}



    	sanei_usb_init();

	i = 0;
	while (g_DeviceID[i].g_nVendor) {
		sanei_usb_find_devices (g_DeviceID[i].g_nVendor, g_DeviceID[i].g_nProduct, check_present);
		i++;
	}
	if (!deviceName) {
		fprintf(stderr, "Device not found!\r\n");
		CU_PRINT("Device not found!");
		exit(EXIT_FAILURE);
	}

	sanei_usb_set_timeout(20000);

	status = sanei_usb_open(deviceName, &usbFd);
	if (status!=SANE_STATUS_GOOD) {
		fprintf(stderr, "sanei_usb_open() error %d\r\n", status);
		CU_PRINT("sanei_usb_open() error %d", status);
		exit(EXIT_FAILURE);
	}

	sanei_usb_reset(usbFd);

	testunitready();
	testunitready();
	inquiry();
	testunitready();

	{
		SANE_Byte userdata_buffer[128]={0};

		readuserdata(userdata_buffer, sizeof(userdata_buffer));

		if (userdata.enableSeparationForceControl < 0)
			BITOFF(userdata_buffer[30], 0x80);
		else if (0 < userdata.enableSeparationForceControl)
			BITON(userdata_buffer[30], 0x80);

		if (userdata.enableManualFeedControl < 0)
			BITOFF(userdata_buffer[31], 0x80);
		else if (0 < userdata.enableManualFeedControl)
			BITON(userdata_buffer[31], 0x80);

		if (userdata.enableManualFeedRollerRotatingTimeControl < 0)
			BITOFF(userdata_buffer[61], 0x80);
		else if (0 < userdata.enableManualFeedRollerRotatingTimeControl)
			BITON(userdata_buffer[61], 0x80);

		if (userdata.enableStapleDetectionControl < 0)
			BITOFF(userdata_buffer[7], 0x80);
		else if (0 < userdata.enableStapleDetectionControl)
			BITON(userdata_buffer[7], 0x80);

		if (userdata.enableLongDocumentControl < 0)
			BITOFF(userdata_buffer[16], 0x80);
		else if (0 < userdata.enableLongDocumentControl)
			BITON(userdata_buffer[16], 0x80);

		if (userdata.enableSlowFeedAtLongDocumentControl < 0)
			BITOFF(userdata_buffer[63], 0x80);
		else if (0 < userdata.enableSlowFeedAtLongDocumentControl)
			BITON(userdata_buffer[63], 0x80);

		if (userdata.enableTrayPositionControl < 0)
			BITOFF(userdata_buffer[66], 0x80);
		else if (0 < userdata.enableTrayPositionControl)
			BITON(userdata_buffer[66], 0x80);

		if (userdata.enableFeedIntervalControl < 0)
			BITOFF(userdata_buffer[64], 0x80);
		else if (0 < userdata.enableFeedIntervalControl)
			BITON(userdata_buffer[64], 0x80);

		if (userdata.enableDustDetetionControl < 0)
			BITOFF(userdata_buffer[28], 0x80);
		else if (0 < userdata.enableDustDetetionControl)
			BITON(userdata_buffer[28], 0x80);

		senduserdata(userdata_buffer, sizeof(userdata_buffer));
	}

	sanei_usb_close(usbFd);

	sanei_usb_exit();

	exit(EXIT_SUCCESS);
}
